/***************************************************************************
 *   Copyright (C) 2004, 2006 by Dominic Rath                              *
 *   Dominic.Rath@gmx.de                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if IS_CYGWIN == 1
#include "windows.h"
#undef ERROR
#endif

//#include "replacements.h"

/* project specific includes */
//#include "log.h"
#include "types.h"
#include "jtag.h"
//#include "configuration.h"
//#include "time_support.h"

/* system includes */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "ftd2xx.h"

#include <sys/time.h>
#include <time.h>

#if IS_WIN32
#include <windows.h>
typedef HINSTANCE DLL_INSTANCE;
#else
#include <dlfcn.h>
typedef void * DLL_INSTANCE;
#endif

/* enable this to debug io latency
 */
#if 0
#define _DEBUG_USB_IO_
#endif

/* enable this to debug communication
 */
#if 0
#define _DEBUG_USB_COMMS_
#endif

int ft2232_uart_read(u8 * buf, int size, u32 * bytes_read);
int ft2232_execute_queue(void);
int ft2232_speed(int speed);
int ft2232_register_commands(struct command_context_s *cmd_ctx);
int ft2232_init(void);
int ft2232_quit(void);
int ft2232_handle_device_desc(char *dev_desc);
int ft2232_handle_serial(char *serial);
int ft2232_handle_layout(char *layout);
//int ft2232_handle_vid_pid_command(struct command_context_s *cmd_ctx, char *cmd, char **args, int argc);
int ft2232_handle_latency_command(struct command_context_s *cmd_ctx,
				  char *cmd, char **args, int argc);

int ft2232_uart_enable(int enable);

char *ft2232_device_desc = NULL;
char *ft2232_serial = NULL;
char *ft2232_layout = NULL;
unsigned char ft2232_latency = 2;

#define MAX_USB_IDS	8
/* vid = pid = 0 marks the end of the list */
static u16 ft2232_vid[MAX_USB_IDS + 1] = { 0x0403, 0 };
static u16 ft2232_pid[MAX_USB_IDS + 1] = { 0xcff8, 0 };

typedef struct ft2232_layout_s {
    char *name;
    int (*init) (void);
    void (*reset) (int trst, int srst);
    void (*blink) (void);
} ft2232_layout_t;

/* init procedures for supported layouts */
int jtagkey_init(void);

/* reset procedures for supported layouts */
void jtagkey_reset(int trst, int srst);
void xdk_reset(int trst, int srst);
void xtag_reset(int trst, int srst);

ft2232_layout_t ft2232_layouts[] = {
    {"jtagkey", jtagkey_init, jtagkey_reset, NULL},
    {"xdk", jtagkey_init, xdk_reset, NULL},
    {"xtag", jtagkey_init, xtag_reset, NULL}, 
    {NULL, NULL, NULL},
};

static u8 nTRST, nTRSTnOE, nSRST, nSRSTnOE;

static ft2232_layout_t *layout;
static u8 low_output = 0x0;
static u8 low_direction = 0x0;
static u8 high_output = 0x0;
static u8 high_direction = 0x0;

static FT_HANDLE ftdih = NULL;
static FT_HANDLE ftdih_uart = NULL;
static int enable_uart = 0;

static u8 *ft2232_buffer = NULL;
static int ft2232_buffer_size = 0;
static int ft2232_read_pointer = 0;
static int ft2232_expect_read = 0;

#define FT2232_BUFFER_SIZE	131072
#define BUFFER_ADD ft2232_buffer[ft2232_buffer_size++]
#define BUFFER_READ ft2232_buffer[ft2232_read_pointer++]

typedef FT_STATUS (*ft_write_fptr)(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesWrittenv);
typedef FT_STATUS (*ft_read_fptr)(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesReturned);
typedef FT_STATUS (*ft_create_device_info_list_fptr)(LPDWORD lpdwNumDevs);
typedef FT_STATUS (*ft_get_device_info_list_fptr)(FT_DEVICE_LIST_INFO_NODE *pDest, LPDWORD lpdwNumDevs);
typedef FT_STATUS (*ft_set_vid_pid_fptr)(DWORD dwVID, DWORD dwPID);
typedef FT_STATUS (*ft_open_ex_fptr)(PVOID pArg1, DWORD Flags, FT_HANDLE *pHandle);
typedef FT_STATUS (*ft_list_devices_fptr)(PVOID pArg1, PVOID pArg2, DWORD Flags);
typedef FT_STATUS (*ft_set_latency_timer_fptr)(FT_HANDLE ftHandle, UCHAR ucLatency);
typedef FT_STATUS (*ft_get_latency_timer_fptr)(FT_HANDLE ftHandle, PUCHAR pucLatency);
typedef FT_STATUS (*ft_set_timeouts_fptr)(FT_HANDLE ftHandle, ULONG ReadTimeout, ULONG WriteTimeout);
typedef FT_STATUS (*ft_set_bit_mode_fptr)(FT_HANDLE ftHandle, UCHAR ucMask, UCHAR ucEnable);
typedef FT_STATUS (*ft_purge_fptr)(FT_HANDLE ftHandle, ULONG Mask);
typedef FT_STATUS (*ft_close_fptr)(FT_HANDLE ftHandle);
typedef FT_STATUS (*ft_ee_program_fptr)(FT_HANDLE ftHandle, PFT_PROGRAM_DATA pData);
typedef FT_STATUS (*ft_set_baud_rate_fptr)(FT_HANDLE ftHandle, DWORD pData);
typedef FT_STATUS (*ft_set_data_characteristics_fptr)(FT_HANDLE ftHandle, UCHAR uWordLength,
		UCHAR uStopBits, UCHAR uParity);
typedef FT_STATUS (*ft_get_status_fptr)(FT_HANDLE ftHandle, LPDWORD lpdwAmountInRxQueue,
		LPDWORD lpdwAmountInTxQueue, LPDWORD lpdwEventStatus);


typedef struct {
  ft_write_fptr write;
  ft_read_fptr read;
  ft_create_device_info_list_fptr create_device_info_list;
  ft_get_device_info_list_fptr get_device_info_list;
  ft_set_vid_pid_fptr set_vid_pid;
  ft_open_ex_fptr open_ex;
  ft_list_devices_fptr list_devices;
  ft_set_latency_timer_fptr set_latency_timer;
  ft_get_latency_timer_fptr get_latency_timer;
  ft_set_timeouts_fptr set_timeouts;
  ft_set_bit_mode_fptr set_bit_mode;
  ft_purge_fptr purge;
  ft_close_fptr close;
  ft_ee_program_fptr ee_program;
  ft_set_baud_rate_fptr set_baud_rate;
  ft_set_data_characteristics_fptr set_data_characteristics;
  ft_get_status_fptr get_status;
} ft_dll_fptrs;

DLL_INSTANCE FT_dll_handle;

ft_dll_fptrs FT_dll;

char *internal_device_list = NULL;
static char ftdi_dll_path[512];

jtag_interface_t ft2232_interface = {
    .name = "ft2232",
    .execute_queue = ft2232_execute_queue,
    .speed = ft2232_speed,
    .register_commands = ft2232_register_commands,
    .init = ft2232_init,
    .quit = ft2232_quit,
};

int ft2232_set_dll_path(char *path) {
  memset(ftdi_dll_path, 0, 512);
  if (path) {
    strcpy(ftdi_dll_path, path);
  }
  return 0;
}

int ft2232_open_dll() {

	if (FT_dll_handle) {
	  return ERROR_OK;
	}

#if IS_WIN32

        if (strlen(ftdi_dll_path)>0) {
          printf("Opening %s\n", ftdi_dll_path);
	  FT_dll_handle = LoadLibrary(ftdi_dll_path);
        } else {
          FT_dll_handle = LoadLibrary("xftd2xx.dll");
        }

	if (!FT_dll_handle) {
	  ERROR("Cannot open FT dll");
	  return ERROR_JTAG_DEVICE_ERROR;
	}

	FT_dll.write = (void*)GetProcAddress(FT_dll_handle, "FT_Write");
	FT_dll.read = (void*)GetProcAddress(FT_dll_handle, "FT_Read");
	FT_dll.create_device_info_list = (void*)GetProcAddress(FT_dll_handle, "FT_CreateDeviceInfoList");
	FT_dll.get_device_info_list = (void*)GetProcAddress(FT_dll_handle, "FT_GetDeviceInfoList");
	//FT_dll.set_vid_pid = (void*)GetProcAddress(FT_dll_handle, "FT_SetVIDPID"); Not Windows
	FT_dll.open_ex = (void*)GetProcAddress(FT_dll_handle, "FT_OpenEx");
	FT_dll.list_devices = (void*)GetProcAddress(FT_dll_handle, "FT_ListDevices");
	FT_dll.set_latency_timer = (void*)GetProcAddress(FT_dll_handle, "FT_SetLatencyTimer");
	FT_dll.get_latency_timer = (void*)GetProcAddress(FT_dll_handle, "FT_GetLatencyTimer");
	FT_dll.set_timeouts = (void*)GetProcAddress(FT_dll_handle, "FT_SetTimeouts");
	FT_dll.set_bit_mode = (void*)GetProcAddress(FT_dll_handle, "FT_SetBitMode");
	FT_dll.purge = (void*)GetProcAddress(FT_dll_handle, "FT_Purge");
	FT_dll.close = (void*)GetProcAddress(FT_dll_handle, "FT_Close");
    FT_dll.ee_program = (void*)GetProcAddress(FT_dll_handle, "FT_EE_Program");
    FT_dll.set_baud_rate = (void*)GetProcAddress(FT_dll_handle, "FT_SetBaudRate");
    FT_dll.set_data_characteristics = (void*)GetProcAddress(FT_dll_handle, "FT_SetDataCharacteristics");
    FT_dll.get_status = (void*)GetProcAddress(FT_dll_handle, "FT_GetStatus");

	//printf("All function pointers open\n");

#else
        if (strlen(ftdi_dll_path)>0) {
          printf("Opening %s\n", ftdi_dll_path);
          FT_dll_handle = dlopen(ftdi_dll_path, RTLD_NOW);
        } else {
          FT_dll_handle = dlopen("xftd2xx.dll", RTLD_NOW);
        }

        if (!FT_dll_handle) {
          ERROR("Cannot open FT so %s", dlerror());
          return ERROR_JTAG_DEVICE_ERROR;
        }

        FT_dll.write = (void*)dlsym(FT_dll_handle, "FT_Write");
        FT_dll.read = (void*)dlsym(FT_dll_handle, "FT_Read");
        FT_dll.create_device_info_list = (void*)dlsym(FT_dll_handle, "FT_CreateDeviceInfoList");
        FT_dll.get_device_info_list = (void*)dlsym(FT_dll_handle, "FT_GetDeviceInfoList");
        FT_dll.set_vid_pid = (void*)dlsym(FT_dll_handle, "FT_SetVIDPID");
        FT_dll.open_ex = (void*)dlsym(FT_dll_handle, "FT_OpenEx");
        FT_dll.list_devices = (void*)dlsym(FT_dll_handle, "FT_ListDevices");
        FT_dll.set_latency_timer = (void*)dlsym(FT_dll_handle, "FT_SetLatencyTimer");
        FT_dll.get_latency_timer = (void*)dlsym(FT_dll_handle, "FT_GetLatencyTimer");
        FT_dll.set_timeouts = (void*)dlsym(FT_dll_handle, "FT_SetTimeouts");
        FT_dll.set_bit_mode = (void*)dlsym(FT_dll_handle, "FT_SetBitMode");
        FT_dll.purge = (void*)dlsym(FT_dll_handle, "FT_Purge");
        FT_dll.close = (void*)dlsym(FT_dll_handle, "FT_Close");
        FT_dll.ee_program = (void*)dlsym(FT_dll_handle, "FT_EE_Program");
        FT_dll.set_baud_rate = (void*)dlsym(FT_dll_handle, "FT_SetBaudRate");
        FT_dll.set_data_characteristics = (void*)dlsym(FT_dll_handle, "FT_SetDataCharacteristics");
        FT_dll.get_status = (void*)dlsym(FT_dll_handle, "FT_GetStatus");
#endif

    if (!FT_dll.write) {
        ERROR("FT_Write function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.read) {
        ERROR("FT_Read function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.create_device_info_list) {
        ERROR("FT_CreateDeviceInfoList function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.get_device_info_list) {
        ERROR("FT_GetDeviceInfoList function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
#if IS_WIN32 == 0
    if (!FT_dll.set_vid_pid) {
        ERROR("FT_SetVIDPID function not found in %s", ftdi_dll_path)
        return ERROR_JTAG_DEVICE_ERROR;
    }
#endif
    if (!FT_dll.open_ex) {
        ERROR("FT_OpenEx function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.list_devices) {
        ERROR("FT_ListDevices function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.set_latency_timer) {
        ERROR("FT_SetLatencyTimer function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.get_latency_timer) {
        ERROR("FT_GetLatencyTimer function not found in %s", ftdi_dll_path)
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.set_timeouts) {
        ERROR("FT_SetTimeouts function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.set_bit_mode) {
        ERROR("FT_SetBitMode function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.purge) {
        ERROR("FT_Purge function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }
    if (!FT_dll.close) {
        ERROR("FT_Close function not found in %s", ftdi_dll_path)
        return ERROR_JTAG_DEVICE_ERROR;
    }
  
    if (!FT_dll.ee_program) {
        ERROR("FT_EE_Program function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }

    if (!FT_dll.set_baud_rate) {
        ERROR("FT_SetBaudRate function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }

    if (!FT_dll.set_data_characteristics) {
        ERROR("FT_SetDataCharacteristics function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }

    if (!FT_dll.get_status) {
        ERROR("FT_GetStatus function not found in %s", ftdi_dll_path);
        return ERROR_JTAG_DEVICE_ERROR;
    }

    return ERROR_OK;
}

int ft2232_write(u8 * buf, int size, u32 * bytes_written)
{
    FT_STATUS status;
    DWORD dw_bytes_written;
    if ((status = FT_dll.write(ftdih, buf, size, &dw_bytes_written)) != FT_OK) {
	*bytes_written = dw_bytes_written;
	ERROR("FT_Write returned: %lu", status);
	return ERROR_JTAG_DEVICE_ERROR;
    } else {
	*bytes_written = dw_bytes_written;
	return ERROR_OK;
    }
}

int ft2232_read(u8 * buf, int size, u32 * bytes_read)
{
    DWORD dw_bytes_read;
    FT_STATUS status;
    int timeout = 5;
    *bytes_read = 0;

    while ((*bytes_read < size) && timeout--) {
	if ((status =
	     FT_dll.read(ftdih, buf + *bytes_read, size - *bytes_read,
		     &dw_bytes_read)) != FT_OK) {
	    *bytes_read = 0;
	    ERROR("FT_Read returned: %lu", status);
	    return ERROR_JTAG_DEVICE_ERROR;
	}
	*bytes_read += dw_bytes_read;
    }

    if (*bytes_read < size) {
	ERROR
	    ("couldn't read the requested number of bytes from FT2232 device (%i < %i)",
	     *bytes_read, size);
	return ERROR_JTAG_DEVICE_ERROR;
    }

    return ERROR_OK;
}

// XMOS
int ft2232_uart_read(u8 * buf, int size, u32 * bytes_read)
{
    FT_STATUS status;
    int timeout = 5;
    int rx_bytes;
    int tx_bytes;
    int event;
    *bytes_read = 0;

//    return ERROR_OK;

    if ((status = FT_dll.get_status(ftdih_uart, &rx_bytes, &tx_bytes, &event)) != FT_OK) {
		ERROR("couldn't read status of UART device");
		return ERROR_JTAG_DEVICE_ERROR;
    }

	size = (rx_bytes > size) ? size : rx_bytes;

	if ((status = FT_dll.read(ftdih_uart, buf, size, bytes_read)) != FT_OK) {
		ERROR("couldn't read from UART buffer");
		return ERROR_JTAG_DEVICE_ERROR;
	}

    return ERROR_OK;
}

int ft2232_speed(int speed)
{
    u8 buf[3];
    int retval;
    u32 bytes_written;

    buf[0] = 0x86;		/* command "set divisor" */
    buf[1] = speed & 0xff;	/* valueL (0=6MHz, 1=3MHz, 2=2.0MHz, ... */
    buf[2] = (speed >> 8) & 0xff;	/* valueH */

    DEBUG("%2.2x %2.2x %2.2x", buf[0], buf[1], buf[2]);
    if (((retval = ft2232_write(buf, 3, &bytes_written)) != ERROR_OK)
	|| (bytes_written != 3)) {
	ERROR("couldn't set FT2232 TCK speed");
	return retval;
    }

    jtag_speed = speed;

    return ERROR_OK;
}

int ft2232_register_commands(struct command_context_s *cmd_ctx)
{
    return ERROR_OK;
}

void ft2232_end_state(enum tap_state state)
{
    if (tap_move_map[state] != -1) {
	end_state = state;
    } else {
	ERROR("BUG: %i is not a valid end state", state);
	exit(-1);
    }
}

void ft2232_read_scan(enum scan_type type, u8 * buffer, int scan_size)
{
    int num_bytes = ((scan_size + 7) / 8);
    int bits_left = scan_size;
    int cur_byte = 0;

    while (num_bytes-- > 1) {
	buffer[cur_byte] = BUFFER_READ;
	cur_byte++;
	bits_left -= 8;
    }

    buffer[cur_byte] = 0x0;

    if (bits_left > 1) {
	buffer[cur_byte] = BUFFER_READ >> 1;
    }

    buffer[cur_byte] =
	(buffer[cur_byte] | ((BUFFER_READ & 0x02) << 6)) >> (8 -
							     bits_left);
}

void ft2232_debug_dump_buffer(void)
{
    int i;
    char line[256];
    char *line_p = line;

    for (i = 0; i < ft2232_buffer_size; i++) {
	//line_p +=
	 //   snprintf(line_p, 256 - (line_p - line), "%2.2x ",
	//	     ft2232_buffer[i]);
	if (i % 16 == 15) {
	    DEBUG("%s", line);
	    line_p = line;
	}
    }

    if (line_p != line)
	DEBUG("%s", line);
}

int ft2232_send_and_recv(jtag_command_t * first, jtag_command_t * last)
{
    jtag_command_t *cmd;
    u8 *buffer;
    int scan_size;
    enum scan_type type;
    int retval;
    u32 bytes_written;
    u32 bytes_read;

#ifdef _DEBUG_USB_IO_
    struct timeval start, inter, inter2, end;
    struct timeval d_inter, d_inter2, d_end;
#endif

#ifdef _DEBUG_USB_COMMS_
    DEBUG("write buffer (size %i):", ft2232_buffer_size);
    ft2232_debug_dump_buffer();
#endif

#ifdef _DEBUG_USB_IO_
    gettimeofday(&start, NULL);
#endif

    if ((retval =
	 ft2232_write(ft2232_buffer, ft2232_buffer_size,
		      &bytes_written)) != ERROR_OK) {
	ERROR("couldn't write MPSSE commands to FT2232");
	exit(-1);
    }
#ifdef _DEBUG_USB_IO_
    gettimeofday(&inter, NULL);
#endif

    if (ft2232_expect_read) {
	int timeout = 100;
	ft2232_buffer_size = 0;

#ifdef _DEBUG_USB_IO_
	gettimeofday(&inter2, NULL);
#endif

	if ((retval =
	     ft2232_read(ft2232_buffer, ft2232_expect_read,
			 &bytes_read)) != ERROR_OK) {
	    ERROR("couldn't read from FT2232");
	    exit(-1);
	}
#ifdef _DEBUG_USB_IO_
	gettimeofday(&end, NULL);

	timeval_subtract(&d_inter, &inter, &start);
	timeval_subtract(&d_inter2, &inter2, &start);
	timeval_subtract(&d_end, &end, &start);

	INFO("inter: %i.%i, inter2: %i.%i end: %i.%i", d_inter.tv_sec,
	     d_inter.tv_usec, d_inter2.tv_sec, d_inter2.tv_usec,
	     d_end.tv_sec, d_end.tv_usec);
#endif

	ft2232_buffer_size = bytes_read;

	if (ft2232_expect_read != ft2232_buffer_size) {
	    ERROR
		("ft2232_expect_read (%i) != ft2232_buffer_size (%i) (%i retries)",
		 ft2232_expect_read, ft2232_buffer_size, 100 - timeout);
	    ft2232_debug_dump_buffer();
	    exit(-1);
	}
#ifdef _DEBUG_USB_COMMS_
	DEBUG("read buffer (%i retries): %i bytes", 100 - timeout,
	      ft2232_buffer_size);
	ft2232_debug_dump_buffer();
#endif
    }

    ft2232_expect_read = 0;
    ft2232_read_pointer = 0;

    /* return ERROR_OK, unless a jtag_read_buffer returns a failed check
     * that wasn't handled by a caller-provided error handler
     */
    retval = ERROR_OK;

    cmd = first;
    while (cmd != last) {
	switch (cmd->type) {
	case JTAG_SCAN:
	    type = jtag_scan_type(cmd->cmd.scan);
	    if (type != SCAN_OUT) {
		scan_size = jtag_scan_size(cmd->cmd.scan);
		buffer = calloc(CEIL(scan_size, 8), 1);
		ft2232_read_scan(type, buffer, scan_size);
		if (jtag_read_buffer(buffer, cmd->cmd.scan) != ERROR_OK) {
		    retval = ERROR_JTAG_QUEUE_FAILED;
		    //free(buffer);
		}
                free(buffer);
	    }
	    break;
	default:
	    break;
	}
	cmd = cmd->next;
    }

    ft2232_buffer_size = 0;

    return retval;
}

void ft2232_add_pathmove(pathmove_command_t * cmd)
{
    int num_states = cmd->num_states;
    u8 tms_byte;
    int state_count;

    state_count = 0;
    while (num_states) {
	tms_byte = 0x0;
	int bit_count = 0;

	int num_states_batch = num_states > 7 ? 7 : num_states;

	/* command "Clock Data to TMS/CS Pin (no Read)" */
	BUFFER_ADD = 0x4b;
	/* number of states remaining */
	BUFFER_ADD = num_states_batch - 1;

	while (num_states_batch--) {
	    if (tap_transitions[cur_state].low == cmd->path[state_count])
		buf_set_u32(&tms_byte, bit_count++, 1, 0x0);
	    else if (tap_transitions[cur_state].high ==
		     cmd->path[state_count])
		buf_set_u32(&tms_byte, bit_count++, 1, 0x1);
	    else {
		ERROR("BUG: %s -> %s isn't a valid TAP transition",
		      tap_state_strings[cur_state],
		      tap_state_strings[cmd->path[state_count]]);
		exit(-1);
	    }

	    cur_state = cmd->path[state_count];
	    state_count++;
	    num_states--;
	}

	BUFFER_ADD = tms_byte;
    }

    end_state = cur_state;
}

void
ft2232_add_scan(int ir_scan, enum scan_type type, u8 * buffer,
		int scan_size)
{
    int num_bytes = (scan_size + 7) / 8;
    int bits_left = scan_size;
    int cur_byte = 0;
    int last_bit;

    if (!((!ir_scan && (cur_state == TAP_SD))
	  || (ir_scan && (cur_state == TAP_SI)))) {
	/* command "Clock Data to TMS/CS Pin (no Read)" */
	BUFFER_ADD = 0x4b;
	/* scan 7 bit */
	BUFFER_ADD = 0x6;
	/* TMS data bits */
	if (ir_scan) {
	    BUFFER_ADD = TAP_MOVE(cur_state, TAP_SI);
	    cur_state = TAP_SI;
	} else {
	    BUFFER_ADD = TAP_MOVE(cur_state, TAP_SD);
	    cur_state = TAP_SD;
	}
	/* DEBUG("added TMS scan (no read)"); */
    }

    /* add command for complete bytes */
    while (num_bytes > 1) {
	int thisrun_bytes;
	if (type == SCAN_IO) {
	    /* Clock Data Bytes In and Out LSB First */
	    BUFFER_ADD = 0x39;
	    /* DEBUG("added TDI bytes (io %i)", num_bytes); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bytes Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x19;
	    /* DEBUG("added TDI bytes (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bytes In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x28;
	    /* DEBUG("added TDI bytes (i %i)", num_bytes); */
	}
	thisrun_bytes = (num_bytes > 65537) ? 65536 : (num_bytes - 1);
	num_bytes -= thisrun_bytes;
	BUFFER_ADD = (thisrun_bytes - 1) & 0xff;
	BUFFER_ADD = ((thisrun_bytes - 1) >> 8) & 0xff;
	if (type != SCAN_IN) {
	    /* add complete bytes */
	    while (thisrun_bytes-- > 0) {
		BUFFER_ADD = buffer[cur_byte];
		cur_byte++;
		bits_left -= 8;
	    }
	} else {		/* (type == SCAN_IN) */

	    bits_left -= 8 * (thisrun_bytes);
	}
    }

    /* the most signifcant bit is scanned during TAP movement */
    if (type != SCAN_IN)
	last_bit = (buffer[cur_byte] >> (bits_left - 1)) & 0x1;
    else
	last_bit = 0;

    /* process remaining bits but the last one */
    if (bits_left > 1) {
	if (type == SCAN_IO) {
	    /* Clock Data Bits In and Out LSB First */
	    BUFFER_ADD = 0x3b;
	    /* DEBUG("added TDI bits (io) %i", bits_left - 1); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bits Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x1b;
	    /* DEBUG("added TDI bits (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bits In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x2a;
	    /* DEBUG("added TDI bits (i %i)", bits_left - 1); */
	}
	BUFFER_ADD = bits_left - 2;
	if (type != SCAN_IN)
	    BUFFER_ADD = buffer[cur_byte];
    }

    if ((ir_scan && (end_state == TAP_SI)) ||
	(!ir_scan && (end_state == TAP_SD))) {
	if (type == SCAN_IO) {
	    /* Clock Data Bits In and Out LSB First */
	    BUFFER_ADD = 0x3b;
	    /* DEBUG("added TDI bits (io) %i", bits_left - 1); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bits Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x1b;
	    /* DEBUG("added TDI bits (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bits In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x2a;
	    /* DEBUG("added TDI bits (i %i)", bits_left - 1); */
	}
	BUFFER_ADD = 0x0;
	BUFFER_ADD = last_bit;
    } else {
	/* move from Shift-IR/DR to end state */
	if (type != SCAN_OUT) {
	    /* Clock Data to TMS/CS Pin with Read */
	    BUFFER_ADD = 0x6b;
	    /* DEBUG("added TMS scan (read)"); */
	} else {
	    /* Clock Data to TMS/CS Pin (no Read) */
	    BUFFER_ADD = 0x4b;
	    /* DEBUG("added TMS scan (no read)"); */
	}
	BUFFER_ADD = 0x6;
	BUFFER_ADD = TAP_MOVE(cur_state, end_state) | (last_bit << 7);
	cur_state = end_state;
    }
}

int
ft2232_large_scan(scan_command_t * cmd, enum scan_type type, u8 * buffer,
		  int scan_size)
{
    int num_bytes = (scan_size + 7) / 8;
    int bits_left = scan_size;
    int cur_byte = 0;
    int last_bit;
    u8 *receive_buffer = malloc(CEIL(scan_size, 8));
    u8 *receive_pointer = receive_buffer;
    u32 bytes_written;
    u32 bytes_read;
    int retval;
    int thisrun_read = 0;

    if (cmd->ir_scan) {
	ERROR("BUG: large IR scans are not supported");
	exit(-1);
    }

    if (cur_state != TAP_SD) {
	/* command "Clock Data to TMS/CS Pin (no Read)" */
	BUFFER_ADD = 0x4b;
	/* scan 7 bit */
	BUFFER_ADD = 0x6;
	/* TMS data bits */
	BUFFER_ADD = TAP_MOVE(cur_state, TAP_SD);
	cur_state = TAP_SD;
    }

    if ((retval =
	 ft2232_write(ft2232_buffer, ft2232_buffer_size,
		      &bytes_written)) != ERROR_OK) {
	ERROR("couldn't write MPSSE commands to FT2232");
	exit(-1);
    }
    DEBUG("ft2232_buffer_size: %i, bytes_written: %i", ft2232_buffer_size,
	  bytes_written);
    ft2232_buffer_size = 0;

    /* add command for complete bytes */
    while (num_bytes > 1) {
	int thisrun_bytes;

	if (type == SCAN_IO) {
	    /* Clock Data Bytes In and Out LSB First */
	    BUFFER_ADD = 0x39;
	    /* DEBUG("added TDI bytes (io %i)", num_bytes); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bytes Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x19;
	    /* DEBUG("added TDI bytes (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bytes In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x28;
	    /* DEBUG("added TDI bytes (i %i)", num_bytes); */
	}
	thisrun_bytes = (num_bytes > 65537) ? 65536 : (num_bytes - 1);
	thisrun_read = thisrun_bytes;
	num_bytes -= thisrun_bytes;
	BUFFER_ADD = (thisrun_bytes - 1) & 0xff;
	BUFFER_ADD = ((thisrun_bytes - 1) >> 8) & 0xff;
	if (type != SCAN_IN) {
	    /* add complete bytes */
	    while (thisrun_bytes-- > 0) {
		BUFFER_ADD = buffer[cur_byte];
		cur_byte++;
		bits_left -= 8;
	    }
	} else {		/* (type == SCAN_IN) */

	    bits_left -= 8 * (thisrun_bytes);
	}

	if ((retval =
	     ft2232_write(ft2232_buffer, ft2232_buffer_size,
			  &bytes_written)) != ERROR_OK) {
	    ERROR("couldn't write MPSSE commands to FT2232");
	    exit(-1);
	}
	DEBUG("ft2232_buffer_size: %i, bytes_written: %i",
	      ft2232_buffer_size, bytes_written);
	ft2232_buffer_size = 0;

	if (type != SCAN_OUT) {
	    if ((retval =
		 ft2232_read(receive_pointer, thisrun_read,
			     &bytes_read)) != ERROR_OK) {
		ERROR("couldn't read from FT2232");
		exit(-1);
	    }
	    DEBUG("thisrun_read: %i, bytes_read: %i", thisrun_read,
		  bytes_read);
	    receive_pointer += bytes_read;
	}
    }

    thisrun_read = 0;

    /* the most signifcant bit is scanned during TAP movement */
    if (type != SCAN_IN)
	last_bit = (buffer[cur_byte] >> (bits_left - 1)) & 0x1;
    else
	last_bit = 0;

    /* process remaining bits but the last one */
    if (bits_left > 1) {
	if (type == SCAN_IO) {
	    /* Clock Data Bits In and Out LSB First */
	    BUFFER_ADD = 0x3b;
	    /* DEBUG("added TDI bits (io) %i", bits_left - 1); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bits Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x1b;
	    /* DEBUG("added TDI bits (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bits In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x2a;
	    /* DEBUG("added TDI bits (i %i)", bits_left - 1); */
	}
	BUFFER_ADD = bits_left - 2;
	if (type != SCAN_IN)
	    BUFFER_ADD = buffer[cur_byte];

	if (type != SCAN_OUT)
	    thisrun_read += 2;
    }

    if (end_state == TAP_SD) {
	if (type == SCAN_IO) {
	    /* Clock Data Bits In and Out LSB First */
	    BUFFER_ADD = 0x3b;
	    /* DEBUG("added TDI bits (io) %i", bits_left - 1); */
	} else if (type == SCAN_OUT) {
	    /* Clock Data Bits Out on -ve Clock Edge LSB First (no Read) */
	    BUFFER_ADD = 0x1b;
	    /* DEBUG("added TDI bits (o)"); */
	} else if (type == SCAN_IN) {
	    /* Clock Data Bits In on +ve Clock Edge LSB First (no Write) */
	    BUFFER_ADD = 0x2a;
	    /* DEBUG("added TDI bits (i %i)", bits_left - 1); */
	}
	BUFFER_ADD = 0x0;
	BUFFER_ADD = last_bit;
    } else {
	/* move from Shift-IR/DR to end state */
	if (type != SCAN_OUT) {
	    /* Clock Data to TMS/CS Pin with Read */
	    BUFFER_ADD = 0x6b;
	    /* DEBUG("added TMS scan (read)"); */
	} else {
	    /* Clock Data to TMS/CS Pin (no Read) */
	    BUFFER_ADD = 0x4b;
	    /* DEBUG("added TMS scan (no read)"); */
	}
	BUFFER_ADD = 0x6;
	BUFFER_ADD = TAP_MOVE(cur_state, end_state) | (last_bit << 7);
	cur_state = end_state;
    }

    if (type != SCAN_OUT)
	thisrun_read += 1;

    if ((retval =
	 ft2232_write(ft2232_buffer, ft2232_buffer_size,
		      &bytes_written)) != ERROR_OK) {
	ERROR("couldn't write MPSSE commands to FT2232");
	exit(-1);
    }
    DEBUG("ft2232_buffer_size: %i, bytes_written: %i", ft2232_buffer_size,
	  bytes_written);
    ft2232_buffer_size = 0;

    if (type != SCAN_OUT) {
	if ((retval =
	     ft2232_read(receive_pointer, thisrun_read,
			 &bytes_read)) != ERROR_OK) {
	    ERROR("couldn't read from FT2232");
	    exit(-1);
	}
	DEBUG("thisrun_read: %i, bytes_read: %i", thisrun_read,
	      bytes_read);
	receive_pointer += bytes_read;
    }

    return ERROR_OK;
}

int ft2232_predict_scan_out(int scan_size, enum scan_type type)
{
    int predicted_size = 3;
    int num_bytes = (scan_size - 1) / 8;

    if (cur_state != TAP_SD)
	predicted_size += 3;

    if (type == SCAN_IN) {	/* only from device to host */
	/* complete bytes */
	predicted_size += (CEIL(num_bytes, 65536)) * 3;
	/* remaining bits - 1 (up to 7) */
	predicted_size += ((scan_size - 1) % 8) ? 2 : 0;
    } else {			/* host to device, or bidirectional */

	/* complete bytes */
	predicted_size += num_bytes + (CEIL(num_bytes, 65536)) * 3;
	/* remaining bits -1 (up to 7) */
	predicted_size += ((scan_size - 1) % 8) ? 3 : 0;
    }

    return predicted_size;
}

int ft2232_predict_scan_in(int scan_size, enum scan_type type)
{
    int predicted_size = 0;

    if (type != SCAN_OUT) {
	/* complete bytes */
	predicted_size +=
	    (CEIL(scan_size, 8) > 1) ? (CEIL(scan_size, 8) - 1) : 0;
	/* remaining bits - 1 */
	predicted_size += ((scan_size - 1) % 8) ? 1 : 0;
	/* last bit (from TMS scan) */
	predicted_size += 1;
    }

    /* DEBUG("scan_size: %i, predicted_size: %i", scan_size, predicted_size); */

    return predicted_size;
}

void jtagkey_reset(int trst, int srst)
{
    if (trst == 1) {
	cur_state = TAP_TLR;
	if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
          high_output &= ~nTRSTnOE;
	else
         high_output &= ~nTRST;
    } else if (trst == 0) {
	if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
	    high_output |= nTRSTnOE;
	else
	    high_output |= nTRST;
    }

    if (srst == 1) {
	if (jtag_reset_config & RESET_SRST_PUSH_PULL)
	    high_output &= ~nSRST;
	else
	    high_output &= ~nSRSTnOE;
    } else if (srst == 0) {
	if (jtag_reset_config & RESET_SRST_PUSH_PULL)
	    high_output |= nSRST;
	else
	    high_output |= nSRSTnOE;
    }

    /* command "set data bits high byte" */
    BUFFER_ADD = 0x82;
    BUFFER_ADD = high_output;
    BUFFER_ADD = high_direction;
    DEBUG
	("trst: %i, srst: %i, high_output: 0x%2.2x, high_direction: 0x%2.2x",
	 trst, srst, high_output, high_direction);
}

void xdk_reset(int trst, int srst)
{
    if (trst == 1) {
        cur_state = TAP_TLR;
        if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
            high_output &= ~nTRSTnOE;
        else
            high_output &= ~nTRST;
    } else if (trst == 0) {
        if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
            high_output |= nTRSTnOE;
        else
            high_output |= nTRST;
    }

    high_output = 0x0;


    if (trst == 0) {
      high_output = high_output | 0x1;
    }

    if (srst == 1) {
      high_output = high_output | 0x2;
    }

    /* command "set data bits high byte" */
    BUFFER_ADD = 0x82;
    BUFFER_ADD = high_output;
    BUFFER_ADD = high_direction;
    DEBUG
        ("trst: %i, srst: %i, high_output: 0x%2.2x, high_direction: 0x%2.2x",
         trst, srst, high_output, high_direction);
}

void xtag_reset(int trst, int srst)
{
    if (trst == 1) {
        cur_state = TAP_TLR;
        if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
            high_output &= ~nTRSTnOE;
        else
            high_output &= ~nTRST;
    } else if (trst == 0) {
        if (jtag_reset_config & RESET_TRST_OPEN_DRAIN)
            high_output |= nTRSTnOE;
        else
            high_output |= nTRST;
    }

    high_output = 0x0;


    if (trst == 0) {
      high_output = high_output | 0x1;
    }

    if (srst == 0) {
      high_output = high_output | 0x2;
    }

    /* command "set data bits high byte" */
    BUFFER_ADD = 0x82;
    BUFFER_ADD = high_output;
    BUFFER_ADD = high_direction;
    DEBUG
        ("trst: %i, srst: %i, high_output: 0x%2.2x, high_direction: 0x%2.2x",
         trst, srst, high_output, high_direction);
}

int ft2232_execute_queue()
{
    jtag_command_t *cmd = jtag_command_queue;	/* currently processed command */
    jtag_command_t *first_unsent = cmd;	/* next command that has to be sent */
    u8 *buffer;
    int scan_size;		/* size of IR or DR scan */
    enum scan_type type;
    int i;
    int predicted_size = 0;
    int require_send = 0;
    int retval;

    /* return ERROR_OK, unless ft2232_send_and_recv reports a failed check
     * that wasn't handled by a caller-provided error handler
     */
    retval = ERROR_OK;

    ft2232_buffer_size = 0;
    ft2232_expect_read = 0;

    /* blink, if the current layout has that feature */
    if (layout->blink)
	layout->blink();

    while (cmd) {
	switch (cmd->type) {
	case JTAG_END_STATE:
	    if (cmd->cmd.end_state->end_state != -1)
		ft2232_end_state(cmd->cmd.end_state->end_state);
	    break;
	case JTAG_RESET:
	    /* only send the maximum buffer size that FT2232C can handle */
	    predicted_size = 3;
	    if (ft2232_buffer_size + predicted_size + 1 >
		FT2232_BUFFER_SIZE) {
		if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		    retval = ERROR_JTAG_QUEUE_FAILED;
		require_send = 0;
		first_unsent = cmd;
	    }

	    layout->reset(cmd->cmd.reset->trst, cmd->cmd.reset->srst);
	    require_send = 1;

#ifdef _DEBUG_JTAG_IO_
	    DEBUG("trst: %i, srst: %i", cmd->cmd.reset->trst,
		  cmd->cmd.reset->srst);
#endif
	    break;
	case JTAG_RUNTEST:
	    /* only send the maximum buffer size that FT2232C can handle */
	    predicted_size = 0;
	    if (cur_state != TAP_RTI)
		predicted_size += 3;
	    predicted_size += 3 * CEIL(cmd->cmd.runtest->num_cycles, 7);
	    if ((cmd->cmd.runtest->end_state != -1)
		&& (cmd->cmd.runtest->end_state != TAP_RTI))
		predicted_size += 3;
	    if ((cmd->cmd.runtest->end_state == -1)
		&& (end_state != TAP_RTI))
		predicted_size += 3;
	    if (ft2232_buffer_size + predicted_size + 1 >
		FT2232_BUFFER_SIZE) {
		if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		    retval = ERROR_JTAG_QUEUE_FAILED;
		require_send = 0;
		first_unsent = cmd;
	    }
	    if (cur_state != TAP_RTI) {
		/* command "Clock Data to TMS/CS Pin (no Read)" */
		BUFFER_ADD = 0x4b;
		/* scan 7 bit */
		BUFFER_ADD = 0x6;
		/* TMS data bits */
		BUFFER_ADD = TAP_MOVE(cur_state, TAP_RTI);
		cur_state = TAP_RTI;
		require_send = 1;
	    }
	    i = cmd->cmd.runtest->num_cycles;
	    while (i > 0) {
		/* command "Clock Data to TMS/CS Pin (no Read)" */
		BUFFER_ADD = 0x4b;
		/* scan 7 bit */
		BUFFER_ADD = (i > 7) ? 6 : (i - 1);
		/* TMS data bits */
		BUFFER_ADD = 0x0;
		cur_state = TAP_RTI;
		i -= (i > 7) ? 7 : i;
		/* DEBUG("added TMS scan (no read)"); */
	    }
	    if (cmd->cmd.runtest->end_state != -1)
		ft2232_end_state(cmd->cmd.runtest->end_state);
	    if (cur_state != end_state) {
		/* command "Clock Data to TMS/CS Pin (no Read)" */
		BUFFER_ADD = 0x4b;
		/* scan 7 bit */
		BUFFER_ADD = 0x6;
		/* TMS data bits */
		BUFFER_ADD = TAP_MOVE(cur_state, end_state);
		cur_state = end_state;
		/* DEBUG("added TMS scan (no read)"); */
	    }
	    require_send = 1;
#ifdef _DEBUG_JTAG_IO_
	    DEBUG("runtest: %i, end in %i", cmd->cmd.runtest->num_cycles,
		  end_state);
#endif
	    break;
	case JTAG_STATEMOVE:
	    /* only send the maximum buffer size that FT2232C can handle */
	    predicted_size = 3;
	    if (ft2232_buffer_size + predicted_size + 1 >
		FT2232_BUFFER_SIZE) {
		if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		    retval = ERROR_JTAG_QUEUE_FAILED;
		require_send = 0;
		first_unsent = cmd;
	    }
	    if (cmd->cmd.statemove->end_state != -1)
		ft2232_end_state(cmd->cmd.statemove->end_state);
	    /* command "Clock Data to TMS/CS Pin (no Read)" */
	    BUFFER_ADD = 0x4b;
	    /* scan 7 bit */
	    BUFFER_ADD = 0x6;
	    /* TMS data bits */
	    BUFFER_ADD = TAP_MOVE(cur_state, end_state);
	    /* DEBUG("added TMS scan (no read)"); */
	    cur_state = end_state;
	    require_send = 1;
#ifdef _DEBUG_JTAG_IO_
	    DEBUG("statemove: %i", end_state);
#endif
	    break;
	case JTAG_PATHMOVE:
	    /* only send the maximum buffer size that FT2232C can handle */
	    predicted_size = 3 * CEIL(cmd->cmd.pathmove->num_states, 7);
	    if (ft2232_buffer_size + predicted_size + 1 >
		FT2232_BUFFER_SIZE) {
		if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		    retval = ERROR_JTAG_QUEUE_FAILED;
		require_send = 0;
		first_unsent = cmd;
	    }
	    ft2232_add_pathmove(cmd->cmd.pathmove);
	    require_send = 1;
#ifdef _DEBUG_JTAG_IO_
	    DEBUG("pathmove: %i states, end in %i",
		  cmd->cmd.pathmove->num_states,
		  cmd->cmd.pathmove->path[cmd->cmd.pathmove->num_states -
					  1]);
#endif
	    break;
	case JTAG_SCAN:
	    scan_size = jtag_build_buffer(cmd->cmd.scan, &buffer);
	    type = jtag_scan_type(cmd->cmd.scan);
	    predicted_size = ft2232_predict_scan_out(scan_size, type);
	    if ((predicted_size + 1) > FT2232_BUFFER_SIZE) {
		DEBUG
		    ("oversized ft2232 scan (predicted_size > FT2232_BUFFER_SIZE)");
		/* unsent commands before this */
		if (first_unsent != cmd)
		    if (ft2232_send_and_recv(first_unsent, cmd) !=
			ERROR_OK)
			retval = ERROR_JTAG_QUEUE_FAILED;

		/* current command */
		if (cmd->cmd.scan->end_state != -1)
		    ft2232_end_state(cmd->cmd.scan->end_state);
		ft2232_large_scan(cmd->cmd.scan, type, buffer, scan_size);
		require_send = 0;
		first_unsent = cmd->next;
		if (buffer)
		    free(buffer);
		break;
	    } else if (ft2232_buffer_size + predicted_size + 1 >
		       FT2232_BUFFER_SIZE) {
		DEBUG
		    ("ft2232 buffer size reached, sending queued commands (first_unsent: %p, cmd: %p)",
		     first_unsent, cmd);
		if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		    retval = ERROR_JTAG_QUEUE_FAILED;
		require_send = 0;
		first_unsent = cmd;
	    }

	    ft2232_expect_read += ft2232_predict_scan_in(scan_size, type);
	    /* DEBUG("new read size: %i", ft2232_expect_read); */
	    if (cmd->cmd.scan->end_state != -1)
		ft2232_end_state(cmd->cmd.scan->end_state);
	    ft2232_add_scan(cmd->cmd.scan->ir_scan, type, buffer,
			    scan_size);
	    require_send = 1;
	    if (buffer)
		free(buffer);
#ifdef _DEBUG_JTAG_IO_
	    DEBUG("%s scan, %i bit, end in %i",
		  (cmd->cmd.scan->ir_scan) ? "IR" : "DR", scan_size,
		  end_state);
#endif
	    break;
	case JTAG_SLEEP:
	    if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
		retval = ERROR_JTAG_QUEUE_FAILED;
	    first_unsent = cmd->next;
	    jtag_sleep(cmd->cmd.sleep->us);
#ifdef _DEBUG_JTAG_IO_
	    DEBUG("sleep %i usec", cmd->cmd.sleep->us);
#endif
	    break;
	default:
	    ERROR("BUG: unknown JTAG command type encountered");
	    exit(-1);
	}
	cmd = cmd->next;
    }

    if (require_send > 0)
	if (ft2232_send_and_recv(first_unsent, cmd) != ERROR_OK)
	    retval = ERROR_JTAG_QUEUE_FAILED;

    return retval;
}

int ft2232_uart_enable(int enable){
    enable_uart = enable;
}

static int ft2232_init_ftd2xx(u16 vid, u16 pid, int more, int *try_more)
{
    FT_STATUS status;
    DWORD openex_flags = 0;
    char *openex_string = NULL;
    u8 latency_timer;
    int i = 0;
    //char *uart_dev;
    char uart_dev[64];
    long dwLoc = 0x222;

    if (ft2232_open_dll() == ERROR_JTAG_DEVICE_ERROR) {
      return ERROR_JTAG_INIT_FAILED;
    }

#if 0
FT_DEVICE_LIST_INFO_NODE *devInfo;
DWORD numDevs;
//
// create the device information list
//

status = FT_dll.create_device_info_list(&numDevs);
if (status == FT_OK) {
  printf("Number of devices is %d\n",numDevs);
}

// allocate storage for list based on numDevs
//
devInfo = (FT_DEVICE_LIST_INFO_NODE*)malloc(sizeof(FT_DEVICE_LIST_INFO_NODE)*numDevs);
//
// get the device information list
//
status = FT_dll.get_device_info_list(devInfo,&numDevs);
if (status == FT_OK) {
  for (i = 0; i < numDevs; i++) {
    printf("Dev %d:\n",i);
    printf(" Flags=0x%x\n",devInfo[i].Flags);
    printf(" Type=0x%x\n",devInfo[i].Type);
    printf(" ID=0x%x\n",devInfo[i].ID);
    printf(" LocId=0x%x\n",devInfo[i].LocId);
    printf(" SerialNumber=%s\n",devInfo[i].SerialNumber);
    printf(" Description=%s\n",devInfo[i].Description);
    printf(" ftHandle=0x%x\n",devInfo[i].ftHandle);
  }
}
#endif

    DEBUG("'ft2232' interface using FTD2XX with '%s' layout (%4.4x:%4.4x)",
	  ft2232_layout, vid, pid);

#if IS_WIN32 == 0
    /* Add non-standard Vid/Pid to the linux driver */
    if ((status = FT_dll.set_vid_pid(vid, pid)) != FT_OK) {
	WARNING("couldn't add %4.4x:%4.4x", vid, pid);
    }
#endif

    if (ft2232_device_desc && ft2232_serial) {
	WARNING
	    ("can't open by device description and serial number, giving precedence to serial");
	ft2232_device_desc = NULL;
    }

    if (ft2232_device_desc) {
	openex_string = ft2232_device_desc;
	openex_flags = FT_OPEN_BY_DESCRIPTION;
    } else if (ft2232_serial) {
	openex_string = ft2232_serial;
	openex_flags = FT_OPEN_BY_SERIAL_NUMBER;
    } else {
	ERROR("neither device description nor serial number specified");
	ERROR
	    ("please add \"ft2232_device_desc <string>\" or \"ft2232_serial <string>\" to your .cfg file");

	return ERROR_JTAG_INIT_FAILED;
    }

    printf("openex_string = %s, open_exflags = 0x%x\n", openex_string, openex_flags);

    if ((status = FT_dll.open_ex(openex_string, openex_flags, &ftdih)) != FT_OK) {
      ERROR("Cannot open device %s [%d]\n", openex_string, status);
    }

    if (enable_uart){
        strcpy(uart_dev, ft2232_serial);
        uart_dev[strlen(uart_dev) - 1] = 'B';
        if ((status = FT_dll.open_ex(uart_dev, openex_flags, &ftdih_uart)) != FT_OK) {
            ERROR("Cannot open device %s [%d]\n", openex_string, status);
            return ERROR_JTAG_INIT_FAILED;
        }

        if ((status = FT_dll.set_baud_rate(ftdih_uart, 115200)) != FT_OK) {
            ERROR("unable to set baud rate: %lu", status);
            return ERROR_JTAG_INIT_FAILED;
        }

        if ((status = FT_dll.set_data_characteristics(ftdih_uart, FT_BITS_8, FT_STOP_BITS_1,
                FT_PARITY_NONE)) != FT_OK) {
            ERROR("unable to set data characteristics: %lu", status);
            return ERROR_JTAG_INIT_FAILED;
        }
    }


#if 0
    strcpy(uart_dev, ft2232_serial);
    uart_dev[strlen(uart_dev) - 1] = 'B';
    printf("opening %s\n", uart_dev);
    if ((status = FT_dll.open_ex(uart_dev, openex_flags, &ftdih)) != FT_OK) {
      ERROR("Cannot open device location 0x222");
    }
#endif


#if 0

    if ((status = FT_dll.open_ex(openex_string, openex_flags, &ftdih)) != FT_OK) {
	DWORD num_devices;

	printf("status != FT_OK\n");

	if (more) {
	    WARNING("unable to open ftdi device (trying more): %lu",
		    status);
	    *try_more = 1;
	    return ERROR_JTAG_INIT_FAILED;
	}

	ERROR("unable to open ftdi device: %lu", status);
	status = FT_dll.list_devices(&num_devices, NULL, FT_LIST_NUMBER_ONLY);

        return ERROR_JTAG_INIT_FAILED;
    }
#endif

    if ((status = FT_dll.set_latency_timer(ftdih, ft2232_latency)) != FT_OK) {
	ERROR("unable to set latency timer: %lu", status);
	return ERROR_JTAG_INIT_FAILED;
    }

    if ((status = FT_dll.get_latency_timer(ftdih, &latency_timer)) != FT_OK) {
	ERROR("unable to get latency timer: %lu", status);
	//return ERROR_JTAG_INIT_FAILED;
    } else {
	DEBUG("current latency timer: %i", latency_timer);
    }

    if ((status = FT_dll.set_timeouts(ftdih, 5000, 5000)) != FT_OK) {
	ERROR("unable to set timeouts: %lu", status);
	return ERROR_JTAG_INIT_FAILED;
    }

    if ((status = FT_dll.set_bit_mode(ftdih, 0x0b, 2)) != FT_OK) {
	ERROR("unable to enable bit i/o mode: %lu", status);
	return ERROR_JTAG_INIT_FAILED;
    }


    if (enable_uart){
        // XMOS - try to open UART device
        strcpy(uart_dev, ft2232_serial);
        uart_dev[strlen(uart_dev) - 1] = 'B';
        printf("opening %s\n", uart_dev);
        if ((status = FT_dll.open_ex(uart_dev, FT_OPEN_BY_SERIAL_NUMBER, &ftdih_uart)) != FT_OK) {
            printf("unable to open uart device: %lu\n", status);
        }

        if ((status = FT_dll.set_baud_rate(ftdih_uart, 115200)) != FT_OK) {
    //    if ((status = FT_dll.set_baud_rate(ftdih_uart, 9600)) != FT_OK) {
        ERROR("unable to set baud rate: %lu", status);
        return ERROR_JTAG_INIT_FAILED;
        }

        if ((status = FT_dll.set_data_characteristics(ftdih_uart, FT_BITS_8, FT_STOP_BITS_1,
                FT_PARITY_NONE)) != FT_OK) {
        ERROR("unable to set data characteristics: %lu", status);
        return ERROR_JTAG_INIT_FAILED;
        }

    }

    return ERROR_OK;
}

static int ft2232_purge_ftd2xx(void)
{
    FT_STATUS status;

    if ((status = FT_dll.purge(ftdih, FT_PURGE_RX | FT_PURGE_TX)) != FT_OK) {
	ERROR("error purging ftd2xx device: %lu", status);
	return ERROR_JTAG_INIT_FAILED;
    }

    if (ftdih_uart != NULL){
        if ((status = FT_dll.purge(ftdih_uart, FT_PURGE_RX | FT_PURGE_TX)) != FT_OK) {
    	ERROR("error purging uart device: %lu", status);
        }
    }

    return ERROR_OK;
}

int ft2232_init(void)
{
    u8 buf[1];
    int retval;
    u32 bytes_written;
    ft2232_layout_t *cur_layout = ft2232_layouts;
    int i;

    if ((ft2232_layout == NULL) || (ft2232_layout[0] == 0)) {
	ft2232_layout = "jtagkey";
	WARNING("No ft2232 layout specified, using default 'jtagkey'");
    }

    while (cur_layout->name) {
	if (strcmp(cur_layout->name, ft2232_layout) == 0) {
	    layout = cur_layout;
	    break;
	}
	cur_layout++;
    }

    if (!layout) {
	ERROR("No matching layout found for %s", ft2232_layout);
	return ERROR_JTAG_INIT_FAILED;
    }

    for (i = 0; 1; i++) {
	/*
	 * "more indicates that there are more IDs to try, so we should
	 * not print an error for an ID mismatch (but for anything
	 * else, we should).
	 *
	 * try_more indicates that the error code returned indicates an
	 * ID mismatch (and nothing else) and that we should proceeed
	 * with the next ID pair.
	 */
	int more = ft2232_vid[i + 1] || ft2232_pid[i + 1];
	int try_more = 0;

	retval = ft2232_init_ftd2xx(ft2232_vid[i], ft2232_pid[i],
				    more, &try_more);
	if (retval >= 0)
	    break;
	if (!more || !try_more)
	    return retval;
    }

    ft2232_buffer_size = 0;
    ft2232_buffer = malloc(FT2232_BUFFER_SIZE);

    if (layout->init() != ERROR_OK)
	return ERROR_JTAG_INIT_FAILED;

    ft2232_speed(jtag_speed);

    buf[0] = 0x85;		/* Disconnect TDI/DO to TDO/DI for Loopback */
    if (((retval = ft2232_write(buf, 1, &bytes_written)) != ERROR_OK)
	|| (bytes_written != 1)) {
	ERROR("couldn't write to FT2232 to disable loopback");
	return ERROR_JTAG_INIT_FAILED;
    }

    printf("ABOUT TO RETURN PURGE\n");

    return ft2232_purge_ftd2xx();

    return ERROR_OK;
}

int jtagkey_init(void)
{
    u8 buf[3];
    u32 bytes_written;

    low_output = 0x08;
    low_direction = 0x1b;

    /* initialize low byte for jtag */
    buf[0] = 0x80;		/* command "set data bits low byte" */
    buf[1] = low_output;	/* value (TMS=1,TCK=0, TDI=0, nOE=0) */
    buf[2] = low_direction;	/* dir (output=1), TCK/TDI/TMS=out, TDO=in, nOE=out */
    DEBUG("%2.2x %2.2x %2.2x", buf[0], buf[1], buf[2]);
 
    printf("INIT JTAG\n");

    if (((ft2232_write(buf, 3, &bytes_written)) != ERROR_OK)
	|| (bytes_written != 3)) {
	ERROR("couldn't initialize FT2232 with 'JTAGkey' layout");
	return ERROR_JTAG_INIT_FAILED;
    }

    if ((strcmp(layout->name, "jtagkey") == 0) ||
        (strcmp(layout->name, "xtag") == 0) ||
        (strcmp(layout->name, "xdk") == 0)) {
	nTRST = 0x01;
	nTRSTnOE = 0x4;
	nSRST = 0x02;
	nSRSTnOE = 0x08;
    } else if ((strcmp(layout->name, "jtagkey_prototype_v1") == 0) ||
	       (strcmp(layout->name, "oocdlink") == 0)) {
	nTRST = 0x02;
	nTRSTnOE = 0x1;
	nSRST = 0x08;
	nSRSTnOE = 0x04;
    } else {
	ERROR("BUG: jtagkey_init called for non jtagkey layout");
	exit(-1);
    }

    high_output = 0x0;
    high_direction = 0x0f;

    if (jtag_reset_config & RESET_TRST_OPEN_DRAIN) {
	high_output |= nTRSTnOE;
	high_output &= ~nTRST;
    } else {
	high_output &= ~nTRSTnOE;
	high_output |= nTRST;
    }
    if (jtag_reset_config & RESET_SRST_PUSH_PULL) {
	high_output &= ~nSRSTnOE;
	high_output |= nSRST;
    } else {
	high_output |= nSRSTnOE;
	high_output &= ~nSRST;
        // XTAG !!!!
        if (strcmp(layout->name, "xtag") == 0) {
          high_output = 0x2;
        }
    }


    /* initialize high port */
    buf[0] = 0x82;		/* command "set data bits high byte" */
    buf[1] = high_output;	/* value */
    buf[2] = high_direction;	/* all outputs (xRST and xRSTnOE) */
    DEBUG("%2.2x %2.2x %2.2x", buf[0], buf[1], buf[2]);

    if (((ft2232_write(buf, 3, &bytes_written)) != ERROR_OK)
	|| (bytes_written != 3)) {
	ERROR("couldn't initialize FT2232 with 'JTAGkey' layout");
	return ERROR_JTAG_INIT_FAILED;
    }

    printf("RETURN\n");

    return ERROR_OK;
}

int ft2232_quit(void)
{
    FT_STATUS status;

    status = FT_dll.close(ftdih);
    free(ft2232_buffer);
    ft2232_buffer = NULL;

    // XMOS
    if (ftdih_uart != NULL){
    	status = FT_dll.close(ftdih_uart);
    }

    return ERROR_OK;
}

int ft2232_handle_device_desc(char *dev_desc)
{
    if (dev_desc != NULL) {
	ft2232_device_desc = strdup(dev_desc);
    } else {
	ERROR
	    ("expected exactly one argument to ft2232_device_desc <description>");
    }

    return ERROR_OK;
}

int ft2232_handle_serial(char *serial)
{
    if (serial != NULL) {
	ft2232_serial = strdup(serial);
    } else {
	ERROR
	    ("expected exactly one argument to ft2232_serial <serial-number>");
    }

    return ERROR_OK;
}

int ft2232_handle_layout(char *layout)
{
    if (layout == NULL)
	return ERROR_OK;

    ft2232_layout = malloc(strlen(layout) + 1);
    strcpy(ft2232_layout, layout);

    return ERROR_OK;
}

#if 0
int ft2232_handle_vid_pid(int vid, int pid)
{
    int i;

    if (argc > MAX_USB_IDS * 2) {
	WARNING("ignoring extra IDs in ft2232_vid_pid "
		"(maximum is %d pairs)", MAX_USB_IDS);
	argc = MAX_USB_IDS * 2;
    }
    if (argc < 2 || (argc & 1)) {
	WARNING("incomplete ft2232_vid_pid configuration directive");
	if (argc < 2)
	    return ERROR_OK;
    }

    for (i = 0; i + 1 < argc; i += 2) {
	ft2232_vid[i >> 1] = strtol(args[i], NULL, 0);
	ft2232_pid[i >> 1] = strtol(args[i + 1], NULL, 0);
    }
    /*
     * Explicitly terminate, in case there are multiples instances of
     * ft2232_vid_pid.
     */
    ft2232_vid[i >> 1] = ft2232_pid[i >> 1] = 0;

    return ERROR_OK;
}
#endif

int ft2232_handle_latency(char *latency)
{
    if (latency != NULL) {
	ft2232_latency = atoi(latency);
    } else {
	ERROR("expected exactly one argument to ft2232_latency <ms>");
    }

    return ERROR_OK;
}


//#define printf printf
int ft2232_list_devices(int *num_devices, char **device_list) {
  FT_DEVICE_LIST_INFO_NODE *devInfo;
  DWORD numDevs = 0;
  FT_STATUS status;
  int i = 0;
  unsigned int dev_list_length = 0;
  char *dev_list_ptr = NULL;
  char dev_id_str[16];

  if (internal_device_list) {
    free(internal_device_list);
  }

  if (ft2232_open_dll() == ERROR_JTAG_DEVICE_ERROR) {
    return ERROR_JTAG_INIT_FAILED;
  }

  printf("%p device_list\n", *device_list);

#if IS_WIN32 == 0
  /* Add non-standard Vid/Pid to the linux driver */
  if ((status = FT_dll.set_vid_pid(ft2232_vid[0], ft2232_pid[0])) != FT_OK) {
      WARNING("couldn't add %4.4x:%4.4x", ft2232_vid[0], ft2232_pid[0]);
  }
#endif

  status = FT_dll.create_device_info_list(&numDevs);
  if (status == FT_OK) {
    printf("Number of devices is %d\n",numDevs);
  }

  *num_devices = numDevs;
  *device_list = NULL;

  if (numDevs == 0) {
    return 0;
  }

  devInfo = (FT_DEVICE_LIST_INFO_NODE*)malloc(sizeof(FT_DEVICE_LIST_INFO_NODE)*numDevs);
  status = FT_dll.get_device_info_list(devInfo,&numDevs);
  if (status == FT_OK) {
    for (i = 0; i < numDevs; i++) {
      printf("Dev %d:\n",i);
      printf(" Flags=0x%x\n",devInfo[i].Flags);
      printf(" Type=0x%x\n",devInfo[i].Type);
      printf(" ID=0x%x\n",devInfo[i].ID);
      printf(" LocId=0x%x\n",devInfo[i].LocId);
      printf(" SerialNumber=%s\n",devInfo[i].SerialNumber);
      printf(" Description=%s\n",devInfo[i].Description);
      printf(" ftHandle=0x%x\n",devInfo[i].ftHandle);
    }
  }


  // Format is Name,Serial,Product:Name,Serial,Product

  for(i = 0; i < numDevs; i++) {
    dev_list_length += strlen(devInfo[i].Description);
    if (devInfo[i].SerialNumber == NULL) {
      dev_list_length += strlen("N/A");
    } else if (strcmp(devInfo[i].SerialNumber, "") == 0) {
      dev_list_length += strlen("N/A");
    } else {
      dev_list_length += strlen(devInfo[i].SerialNumber);
    }
    //itoa(devInfo[i].ID, dev_id_str, 16);
    sprintf(dev_id_str, "%x", devInfo[i].ID);
    dev_list_length += strlen(dev_id_str);
  }
  dev_list_length += 2*numDevs;  // ,
  dev_list_length += numDevs;    // :
  dev_list_length += 1;          // \0

  printf("%p device_list\n", *device_list);

  printf("malloc device list of %d bytes\n", dev_list_length);

  internal_device_list = malloc(dev_list_length);
  memset(internal_device_list, 0, dev_list_length);

  printf("%p device_list\n", *device_list);

  dev_list_ptr = internal_device_list;

  for(i = 0; i < numDevs; i++) {
     strcpy(dev_list_ptr, devInfo[i].Description);
     printf("dev_list_ptr = %s\n", dev_list_ptr);
     dev_list_ptr += strlen(devInfo[i].Description);
     *dev_list_ptr = ',';
     dev_list_ptr++;
     if (devInfo[i].SerialNumber == NULL) {
       strcpy(dev_list_ptr, "N/A");
       printf("dev_list_ptr = %s\n", dev_list_ptr);
       dev_list_ptr += strlen("N/A");
     } else if (strcmp(devInfo[i].SerialNumber, "") == 0) {
       strcpy(dev_list_ptr, "N/A");
       printf("dev_list_ptr = %s\n", dev_list_ptr);
       dev_list_ptr += strlen("N/A");
     } else {
       strcpy(dev_list_ptr, devInfo[i].SerialNumber);
       printf("dev_list_ptr = %s\n", dev_list_ptr);
       dev_list_ptr += strlen(devInfo[i].SerialNumber);
     }
     *dev_list_ptr = ',';
     dev_list_ptr++;
     //itoa(devInfo[i].ID, dev_id_str, 16);
     sprintf(dev_id_str, "%x", devInfo[i].ID);
     strcpy(dev_list_ptr, dev_id_str);
     printf("dev_list_ptr = %s\n", dev_list_ptr);
     dev_list_ptr += strlen(dev_id_str);
     if (i < numDevs-1) {
       *dev_list_ptr = ';';
       dev_list_ptr++;
     }
  }

  *dev_list_ptr = '\0';

  //free(devInfo);

  printf("dev_list = %s\n", *device_list);

  printf("Returning ERROR_OK\n");

  *device_list = internal_device_list;

  return ERROR_OK;
}

int ft2232_program_eeprom(void *eeprom_data) {
  PFT_PROGRAM_DATA data = (PFT_PROGRAM_DATA)eeprom_data;
  FT_STATUS status = ERROR_OK;
  if (ftdih) {
    status = FT_dll.ee_program(ftdih, data);
  }
  if (status == ERROR_OK) {
    return 0;
  } else {
    return 1;
  }
}



//#define printf 0 &&

