#include "jtag.h"
#include <stdlib.h>
#include <string.h>
#include "xmos_jtag.h"

extern int jtag_set_speed(int speed);
extern int jtag_simple_irscan(unsigned char *irval, unsigned int bits, unsigned char *output, int execute);
extern int jtag_simple_drscan(unsigned char *drval, unsigned int bits, unsigned char *output, int execute);
extern int jtag_add_reset(int trst, int srst);
extern int ft2232_list_devices(int *num_devices, char **device_list);
extern int ft2232_set_dll_path(char *path);
extern int ft2232_program_eeprom(void *eeprom_data);
extern int ft2232_quit(void);
extern int ft2232_speed(int speed);
int ft2232_uart_read(u8 * buf, int size, u32 * bytes_read);
int ft2232_uart_enable(int enable);

static int xmos_jtag_execute_state = 1;
static int xmos_jtag_device_type = 0;
static int xmos_jtag_connected = 0;

int xmos_jtag_init (int device_type, char *args, int enable_uart) {
  /* initialize commandline interface */
  char *test[16];
  command_context_t *cmd_ctx, *cfg_cmd_ctx;

  cmd_ctx = command_init();

  test[0] = malloc(strlen("ft2232")+1);
  strcpy(test[0], "ft2232");
  handle_interface_command(cmd_ctx, NULL, test, 1);

  if (args) {
    if (strstr(args, "S=")) {
      ft2232_handle_serial(args+strlen("S="));
    } else if (strstr(args, "D=")) {
      ft2232_handle_device_desc(args+strlen("D="));
    } else {
      ft2232_handle_device_desc("Amontec JTAGkey A");
    }
  } else {
    ft2232_handle_device_desc("Amontec JTAGkey A");
  }

  // Layout type determines reset type
  if (device_type == XMOS_JTAG_DEVICE_JTAGKEY) {
    ft2232_handle_layout("jtagkey");
  } else if (device_type == XMOS_JTAG_DEVICE_XDK || device_type == XMOS_JTAG_DEVICE_XC5) {
    ft2232_handle_layout("xdk");
  } else if (device_type == XMOS_JTAG_DEVICE_XTAG) {
    ft2232_handle_layout("xtag");
  } else {
    ft2232_handle_layout("jtagkey");
  }

  ft2232_uart_enable(enable_uart);

  //handle_jtag_speed("0", NULL);
  jtag_interface_init(cmd_ctx);
  jtag_init(cmd_ctx);

  xmos_jtag_device_type = device_type;
  xmos_jtag_connected = 1;

  return 0;
}

int xmos_jtag_deinit(void) {
  ft2232_quit();

  xmos_jtag_device_type = 0;
  xmos_jtag_connected = 0;
  
  return 0;
}

int xmos_jtag_speed(short divisor) {
  return 0;
}

int xmos_jtag_reset(int type) {
  printf("RESET type %d\n", type);

  if (xmos_jtag_device_type == XMOS_JTAG_DEVICE_XTAG && type == XMOS_JTAG_RESET_TRST_SRST) {
    // Overide default full reset with reset into JTAG MODE
    type = XMOS_JTAG_RESET_TRST_SRST_JTAG;
  }
 
  if (xmos_jtag_device_type == XMOS_JTAG_DEVICE_XC5 && type == XMOS_JTAG_RESET_TRST_SRST) {
    // Overide default full reset with reset into JTAG MODE
    type = XMOS_JTAG_RESET_TRST_SRST_JTAG;
  }
 
  if (type == XMOS_JTAG_RESET_TRST_SRST) {
    jtag_add_reset(1, 1);
    jtag_execute_queue();
    jtag_add_reset(0, 0);
    jtag_execute_queue();
  }
  if (type == XMOS_JTAG_RESET_TRST) {
    jtag_add_reset(1, 0);
    jtag_execute_queue();
    jtag_add_reset(0, 0);
    jtag_execute_queue();
  }
  if (type == XMOS_JTAG_RESET_TRST_SRST_JTAG) {
    jtag_add_reset(1, 1);
    jtag_execute_queue();
    jtag_add_reset(1, 0);
    jtag_execute_queue();
    jtag_add_reset(0, 0);
    jtag_execute_queue();
  }
  //getchar();
  return 0;
#define printf 0 &&
}

int xmos_jtag_irscan(unsigned char *bit_data, unsigned int bits) {
  jtag_simple_irscan(bit_data,bits,NULL, xmos_jtag_execute_state);
  return 0;
}

int xmos_jtag_drscan(unsigned char *bit_data, unsigned int bits) {
  jtag_simple_drscan(bit_data,bits,bit_data, xmos_jtag_execute_state);
  return 0;
}

int xmos_jtag_set_execute(int execute) {
  if (!xmos_jtag_execute_state)
    jtag_execute_queue();
  xmos_jtag_execute_state = execute;
  return 0;
}

int xmos_jtag_get_execute(void) {
  return xmos_jtag_execute_state;
}

int xmos_jtag_statemove(int state) {
  return 0;
}

int xmos_jtag_read_uart(unsigned char *buf, int size, int *read){
	int bytes_read;
	int res = ft2232_uart_read(buf, size, read);
	return res == ERROR_OK;
}

int xmos_jtag_list_devices(int *num_devices, char **device_list) {
  ft2232_list_devices(num_devices, device_list);
  return 0;
}

// Command interface for non-standard requests
int xmos_jtag_command(char *command, void *args) {
  printf("command = %s\n", command);
  if (strncmp(command, "ftdilibname=", strlen("ftdilibname=")) == 0) {
    printf("FTDI DLL name = %s\n", command+strlen("ftdilibname="));
    return ft2232_set_dll_path(command+strlen("ftdilibname="));
  }
  if (strcmp(command, "ftdieewrite") == 0) {
    return ft2232_program_eeprom(args);
  }
  if (strcmp(command, "jtagspeed") == 0) {
    if (args) {
      int jtag_speed_divisor = atoi(args);
      printf("Setting speed to %d\n", jtag_speed_divisor);
      jtag_set_speed(jtag_speed_divisor);
      if (xmos_jtag_connected) {
        ft2232_speed(jtag_speed_divisor);
      }
    }
  }
}
  


